import abc
import typing
from http import HTTPStatus as _HTTPStatus
from urllib.parse import urlparse as _urlparse

import requests

from .utils import types as _types

__all__ = [
    "BitrixAPIAccessDenied",
    "BitrixAPIAllowedOnlyIntranetUser",
    "BitrixAPIBadRequest",
    "BitrixAPIError",
    "BitrixAPIErrorBatchLengthExceeded",
    "BitrixAPIErrorBatchMethodNotAllowed",
    "BitrixAPIErrorManifestIsNotAvailable",
    "BitrixAPIErrorOAuth",
    "BitrixAPIErrorUnexpectedAnswer",
    "BitrixAPIExpiredToken",
    "BitrixAPIForbidden",
    "BitrixAPIInsufficientScope",
    "BitrixAPIInternalServerError",
    "BitrixAPIInvalidArgValue",
    "BitrixAPIInvalidCredentials",
    "BitrixAPIInvalidRequest",
    "BitrixAPIMethodConfirmDenied",
    "BitrixAPIMethodConfirmWaiting",
    "BitrixAPIMethodNotAllowed",
    "BitrixAPINoAuthFound",
    "BitrixAPINotFound",
    "BitrixAPIOverloadLimit",
    "BitrixAPIQueryLimitExceeded",
    "BitrixAPIServiceUnavailable",
    "BitrixAPIUnauthorized",
    "BitrixAPIUserAccessError",
    "BitrixOAuthException",
    "BitrixOAuthInsufficientScope",
    "BitrixOAuthInvalidClient",
    "BitrixOAuthInvalidGrant",
    "BitrixOAuthInvalidRequest",
    "BitrixOAuthInvalidScope",
    "BitrixOAuthRequestError",
    "BitrixOAuthRequestTimeout",
    "BitrixOauthWrongClient",
    "BitrixRequestError",
    "BitrixRequestTimeout",
    "BitrixResponse302JSONDecodeError",
    "BitrixResponse500JSONDecodeError",
    "BitrixResponseJSONDecodeError",
    "BitrixSDKException",
    "BitrixValidationError",
]


class _HTTPResponse(abc.ABC):
    """"""

    STATUS_CODE: _HTTPStatus = NotImplemented

    response: requests.Response

    @property
    def status_code(self) -> int:
        """"""
        return self.response.status_code


class _HTTPResponseOK(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.OK


class _HTTPResponseFound(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.FOUND


class _HTTPResponseBadRequest(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.BAD_REQUEST


class _HTTPResponseUnauthorized(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.UNAUTHORIZED


class _HTTPResponseForbidden(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.FORBIDDEN


class _HTTPResponseNotFound(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.NOT_FOUND


class _HTTPResponseMethodNotAllowed(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.METHOD_NOT_ALLOWED


class _HTTPResponseInternalError(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.INTERNAL_SERVER_ERROR


class _HTTPResponseServiceUnavailable(_HTTPResponse):
    """"""
    STATUS_CODE = _HTTPStatus.SERVICE_UNAVAILABLE


class BitrixSDKException(Exception):
    """BaseEntity class for all bitrix API exceptions."""

    __slots__ = ("message",)

    def __init__(self, message: typing.Text, *args):
        super().__init__(message, *args)
        self.message = message

    def __str__(self) -> typing.Text:
        return self.message


class BitrixOAuthException(BitrixSDKException):
    """"""


class BitrixValidationError(BitrixSDKException):
    """"""


class BitrixRequestError(BitrixSDKException):
    """A Connection error occurred."""

    __slots__ = ("original_error",)

    def __init__(self, original_error: Exception, *args):
        super().__init__(f"{self.__class__.__name__}: {original_error}", original_error, *args)
        self.original_error = original_error


class BitrixOAuthRequestError(BitrixRequestError, BitrixOAuthException):
    """"""


class BitrixRequestTimeout(BitrixRequestError):
    """"""

    __slots__ = ("timeout",)

    def __init__(self, original_error: Exception, timeout: int):
        super().__init__(original_error, timeout)
        self.timeout = timeout


class BitrixOAuthRequestTimeout(BitrixRequestTimeout, BitrixOAuthException):
    """"""


class BitrixResponseJSONDecodeError(BitrixRequestError, _HTTPResponse):
    """"""

    __slots__ = ("response",)

    def __init__(self, original_error: Exception, response: requests.Response):
        super().__init__(original_error, response)
        self.response = response


class BitrixResponse302JSONDecodeError(BitrixResponseJSONDecodeError, _HTTPResponseFound):
    """"""

    @property
    def redirect_url(self) -> typing.Optional[typing.Text]:
        """"""
        return self.response.headers.get("Location")

    @property
    def new_domain(self) -> typing.Optional[typing.Text]:
        """"""
        redirect_url = self.redirect_url
        return redirect_url and _urlparse(redirect_url).hostname


class BitrixResponse500JSONDecodeError(BitrixResponseJSONDecodeError, _HTTPResponseInternalError):
    """"""


class BitrixAPIError(BitrixSDKException, _HTTPResponse):
    """"""

    ERROR: typing.Text = NotImplemented

    __slots__ = ("json_response", "response")

    def __init__(self, json_response: _types.JSONDict, response: requests.Response):
        message = json_response.get("error_description", f"{self.__class__.__name__}: {response.text}")
        super().__init__(message, json_response, response)
        self.json_response = json_response
        self.response = response

    @property
    def error(self) -> typing.Text:
        """"""
        return self.json_response.get("error", "")

    @property
    def error_description(self) -> typing.Text:
        """"""
        return self.json_response.get("error_description", "")


# Exceptions by status code

class BitrixAPIBadRequest(BitrixAPIError, _HTTPResponseBadRequest):
    """Bad Request."""


class BitrixAPIUnauthorized(BitrixAPIError, _HTTPResponseUnauthorized):
    """Unauthorized."""


class BitrixAPIForbidden(BitrixAPIError, _HTTPResponseForbidden):
    """Forbidden."""


class BitrixAPINotFound(BitrixAPIError, _HTTPResponseNotFound):
    """Not Found."""
    ERROR = "NOT_FOUND"


class BitrixAPIMethodNotAllowed(BitrixAPIError, _HTTPResponseMethodNotAllowed):
    """Method Not Allowed."""


class BitrixAPIInternalServerError(BitrixAPIError, _HTTPResponseInternalError):
    """Internal server error."""
    ERROR = "INTERNAL_SERVER_ERROR"


class BitrixAPIServiceUnavailable(BitrixAPIError, _HTTPResponseServiceUnavailable):
    """Service Unavailable."""


# Exceptions by error

# 200

class BitrixOauthWrongClient(BitrixAPIError, BitrixOAuthException, _HTTPResponseOK):
    """Wrong client"""
    ERROR = "WRONG_CLIENT"


# 400

class BitrixAPIErrorBatchLengthExceeded(BitrixAPIBadRequest):
    """Max batch length exceeded."""
    ERROR = "ERROR_BATCH_LENGTH_EXCEEDED"


class BitrixAPIInvalidArgValue(BitrixAPIBadRequest):
    """"""
    ERROR = "INVALID_ARG_VALUE"


class BitrixAPIInvalidRequest(BitrixAPIBadRequest):
    """Https required."""
    ERROR = "INVALID_REQUEST"


class BitrixOAuthInvalidRequest(BitrixAPIInvalidRequest, BitrixOAuthException):
    """An incorrectly formatted authorization requests was provided"""


class BitrixOAuthInvalidClient(BitrixAPIBadRequest, BitrixOAuthException):
    """Invalid client data was provided. The application may not be installed in Bitrix24"""
    ERROR = "INVALID_CLIENT"


class BitrixOAuthInvalidGrant(BitrixAPIBadRequest, BitrixOAuthException):
    """Invalid authorization tokens were provided when obtaining access_token. This occurs during renewal or initial acquisition"""
    ERROR = "INVALID_GRANT"


# 401

class BitrixAPIErrorOAuth(BitrixAPIUnauthorized):
    """Application not installed."""
    ERROR = "ERROR_OAUTH"


class BitrixAPIExpiredToken(BitrixAPIUnauthorized):
    """The access token provided has expired."""
    ERROR = "EXPIRED_TOKEN"


class BitrixAPIMethodConfirmWaiting(BitrixAPIUnauthorized):
    """Waiting for confirmation."""
    ERROR = "METHOD_CONFIRM_WAITING"


class BitrixAPINoAuthFound(BitrixAPIUnauthorized):
    """Wrong authorization data."""
    ERROR = "NO_AUTH_FOUND"


# 403

class BitrixAPIAccessDenied(BitrixAPIForbidden):
    """REST API is available only on commercial plans."""
    ERROR = "ACCESS_DENIED"


class BitrixAPIAllowedOnlyIntranetUser(BitrixAPIForbidden):
    """"""
    ERROR = "ALLOWED_ONLY_INTRANET_USER"


class BitrixAPIInsufficientScope(BitrixAPIForbidden):
    """The request requires higher privileges than provided by the webhook token."""
    ERROR = "INSUFFICIENT_SCOPE"


class BitrixAPIInvalidCredentials(BitrixAPIForbidden):
    """Invalid requests credentials."""
    ERROR = "INVALID_CREDENTIALS"


class BitrixAPIMethodConfirmDenied(BitrixAPIForbidden):
    """Method call denied."""
    ERROR = "METHOD_CONFIRM_DENIED"


class BitrixAPIUserAccessError(BitrixAPIForbidden):
    """The user does not have acfcess to the application."""
    ERROR = "USER_ACCESS_ERROR"


class BitrixOAuthInvalidScope(BitrixAPIForbidden, BitrixOAuthException):
    """Access permissions requested exceed those specified in the application card"""
    ERROR = "INVALID_SCOPE"


class BitrixOAuthInsufficientScope(BitrixAPIInsufficientScope, BitrixOAuthException):
    """Access permissions requested exceed those specified in the application card"""


# 404

class BitrixAPIErrorManifestIsNotAvailable(BitrixAPINotFound):
    """Manifest is not available"""
    ERROR = "ERROR_MANIFEST_IS_NOT_AVAILABLE"


# 405

class BitrixAPIErrorBatchMethodNotAllowed(BitrixAPIMethodNotAllowed):
    """Method is not allowed for batch usage."""
    ERROR = "ERROR_BATCH_METHOD_NOT_ALLOWED"


# 500

class BitrixAPIErrorUnexpectedAnswer(BitrixAPIInternalServerError):
    """Server returned an unexpected responses."""
    ERROR = "ERROR_UNEXPECTED_ANSWER"


# 503

class BitrixAPIOverloadLimit(BitrixAPIServiceUnavailable):
    """REST API is blocked due to overload."""
    ERROR = "OVERLOAD_LIMIT"


class BitrixAPIQueryLimitExceeded(BitrixAPIServiceUnavailable):
    """Too many requests."""
    ERROR = "QUERY_LIMIT_EXCEEDED"



from typing import Optional

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Type",
]


class Type(BaseCRM):
    """The methods provide capabilities for working with SPAs.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/index.html#methods-for-working-with-spas
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get custom fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-fields.html

        This method retrieves information about the custom fields of the smart process settings.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new custom type (SPA).

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-add.html

        This method creates a new SPA.

        Args:
            fields: Field values for adding a new SPA;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get custom type by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-get.html

        This method retrieves information about the SPA with the identifier ID.

        Args:
            bitrix_id: Identifier of the SPA;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def get_by_entity_type_id(
            self,
            entity_type_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get smart process type by entityTypeId.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-get-by-entity-type-id.html

        This method retrieves information about the SPA with the smart process type identifier entityTypeId.

        Args:
            entity_type_id: Identifier of the smart process type;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get_by_entity_type_id,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of custom types.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-list.html

        This method retrieves a list of smart process settings.

        Args:
            filter: Object format:
                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        return self._list(
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update user type.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-update.html

        This method updates an existing SPA by its identifier ID.

        Args:
            bitrix_id: Identifier of the SPA;

            fields: Field values for adding a new SPA;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete custom type.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-object-types/crm-type-delete.html

        This method deletes an existing SPA by its ID. You can only delete an SPA if there are no associated CRM entities.

        Args:
            bitrix_id: Identifier of the SPA;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)



from typing import Optional, Protocol, Text

from ...utils.types import JSONDict, Timeout


class BitrixTokenProtocol(Protocol):
    """"""

    domain: Text
    is_webhook: bool
    auth_token: Text
    refresh_token: Optional[Text]

    def call_method(
            self,
            api_method: Text,
            params: Optional[JSONDict] = None,
            *,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict: ...



import os
import time
from abc import ABC, abstractmethod
from http import HTTPStatus
from typing import Dict, Final, Optional, Text, Tuple

import requests
from uuid6 import uuid7

from ..._config import Config
from ..._constants import TEXT_PYTHON_VERSION
from ...utils.types import DefaultTimeout, JSONDict, Number, Timeout
from ...version import SDK_VERSION
from ._utils import parse_response


class BaseRequester(ABC):
    """"""

    _DEFAULT_REQUEST_ID_HEADER_FIELD_NAME: Final[Text] = "X-Request-ID"
    _SDK_VERSION: Final[Text] = SDK_VERSION
    _SDK_USER_AGENT: Final[Text] = "b24-python-sdk-vendor"
    _TEXT_PYTHON_VERSION: Final[Text] = TEXT_PYTHON_VERSION

    _KEY_NAME_VARIANTS: Final[Tuple[Text, ...]] = (
        "REQUEST_ID",
        "HTTP_X_REQUEST_ID",
        "UNIQUE_ID",
    )

    __slots__ = ("_config", "_initial_retry_delay", "_max_retries", "_retries_remaining", "_retry_delay_increment", "_timeout")

    _config: Config
    _initial_retry_delay: Number
    _max_retries: int
    _retries_remaining: int
    _retry_delay_increment: Number
    _timeout: DefaultTimeout

    def __init__(
            self,
            *,
            timeout: Timeout = None,
            max_retries: Optional[int] = None,
            initial_retry_delay: Optional[Number] = None,
            retry_delay_increment: Optional[Number] = None,
    ):
        self._config = Config()
        self._timeout = timeout or self._config.default_timeout
        self._max_retries = max_retries or self._config.default_max_retries
        self._retries_remaining = self._max_retries
        self._initial_retry_delay = initial_retry_delay or self._config.default_initial_retry_delay
        self._retry_delay_increment = retry_delay_increment or self._config.default_retry_delay_increment

    @property
    @abstractmethod
    def _headers(self) -> Dict:
        """"""
        raise NotImplementedError

    def _get_default_headers(self) -> Dict[Text, Text]:
        """"""
        return {
            "Accept": "application/json",
            "Accept-Charset": "utf-8",
            "User-Agent": f"{self._SDK_USER_AGENT}-v-{self._SDK_VERSION}-python-{self._TEXT_PYTHON_VERSION}",
            "X-BITRIX24-PYTHON-SDK-PYTHON-VERSION": self._TEXT_PYTHON_VERSION,
            "X-BITRIX24-PYTHON-SDK-VERSION": self._SDK_VERSION,
            self._DEFAULT_REQUEST_ID_HEADER_FIELD_NAME: self.get_request_id(),
        }

    @abstractmethod
    def _request(self, *args, **kwargs) -> requests.Response:
        """"""
        raise NotImplementedError

    @classmethod
    def _parse_response(cls, response: requests.Response) -> JSONDict:
        """"""
        return parse_response(response)

    @property
    def _retry_timeout(self) -> float:
        """
        Calculates timeout between retries based on amount of retries used

        Returns:
            time to wait before next requests in seconds
        """
        used_retries = self._max_retries - self._retries_remaining
        return self._initial_retry_delay + used_retries * self._retry_delay_increment

    def _request_with_retries(self, *args, **kwargs) -> requests.Response:
        """"""

        self._retries_remaining = self._max_retries

        while self._retries_remaining > 0:

            self._retries_remaining -= 1
            response = self._request(*args, **kwargs)

            if response.status_code == HTTPStatus.SERVICE_UNAVAILABLE and self._retries_remaining:
                retry_count = self._max_retries - self._retries_remaining

                self._config.logger.warning(
                    "Service unavailable!",
                    context=dict(
                        url=response.url,
                        retry_count=retry_count,
                        max_retries=self._max_retries,
                        retries_remaining=self._retries_remaining,
                    ),
                )
                self._config.logger.info(
                    "Sleep before retry",
                    context=dict(
                        sleep_time=self._retry_timeout,
                    ),
                )
                time.sleep(self._retry_timeout)
                continue

            return response

        raise RuntimeError(
            f"Request failed after {self._max_retries} attempts. "
            "No valid responses was received from the server.",
        )

    def _find_exists(self) -> Optional[Text]:
        """Find an existing requests id in environment variables."""

        for key in self._KEY_NAME_VARIANTS:
            request_id = os.environ.get(key)
            if request_id:
                return request_id

        return None

    @staticmethod
    def _generate() -> Text:
        """Generate a new UUIDv7 requests ID."""
        return str(uuid7())

    def get_request_id(self) -> Text:
        """Get existing requests id or generate a new one."""

        request_id = self._find_exists()

        if request_id is None:
            request_id = self._generate()

        return request_id



from dataclasses import asdict, dataclass
from typing import Dict, Optional

from ..._constants import PYTHON_VERSION as _PV
from ...utils.types import B24APIResult, JSONDict
from .bitrix_api_time_response import BitrixAPITimeResponse

_DATACLASS_KWARGS = {"repr": False, "eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class BitrixAPIResponse:
    """"""

    result: B24APIResult
    time: BitrixAPITimeResponse
    next: Optional[int]
    total: Optional[int]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"result={self.result}, "
            f"time={self.time}, "
            f"next={self.next}, "
            f"total={self.total})"
        )

    @classmethod
    def from_dict(cls, json_response: JSONDict) -> "BitrixAPIResponse":
        return cls(
            result=json_response["result"],
            time=BitrixAPITimeResponse.from_dict(json_response["time"]),
            next=json_response.get("next"),
            total=json_response.get("total"),
        )

    def to_dict(self) -> Dict:
        return asdict(self)



from functools import cached_property
from typing import Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM
from .base import Base
from .localizations import Localizations

__all__ = [
    "Currency",
]


class Currency(BaseCRM):
    """The methods provide capabilities for managing currencies.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/index.html
    """

    @cached_property
    def base(self) -> Base:
        """"""
        return Base(self)

    @cached_property
    def localizations(self) -> Localizations:
        """"""
        return Localizations(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get currency fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-fields.html

        The method retrieves the description of currency fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add currency

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-add.html

        The method adds a new currency.

        Args:
            fields: Object format:

                {
                    CURRENCY: 'value',

                    BASE: 'value',

                    AMOUNT_CNT: 'value',

                    AMOUNT: 'value'

                    SORT: 'value'

                    LANG: 'value'
                };

            timeout: timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields=fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get currency by ID

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-get.html

        The method retrieves currency data by its symbolic identifier according to ISO 4217.

        Args:
            bitrix_id: Symbolic identifier of the currency;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of currencies.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-list.html

        The method retrieves the list of currencies created in the account.

        Args:
            order: Object format:

                {
                    field_1: order_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted,

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(order=order, timeout=timeout)

    @type_checker
    def update(
            self,
            bitrix_id: Text,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update currency.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-update.html

        This method updates an existing currency.

        Args:
            bitrix_id: Symbolic identifier;

            fields: Object format:

                {
                    CURRENCY: 'value',

                    BASE: 'value',

                    AMOUNT_CNT: 'value',

                    AMOUNT: 'value'

                    SORT: 'value'

                    LANG: 'value'
                };

            timeout: timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ID": bitrix_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete currency.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/crm-currency-delete.html

        The method deletes a currency.

        Args:
            bitrix_id: Currency symbolic identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Dict, Final, List, Mapping, Optional, Sequence, Text, Tuple, Union, overload

from ..._constants import MAX_BATCH_SIZE
from ...utils.types import B24BatchMethods, B24BatchMethodTuple, JSONDict, JSONList, Key, Timeout
from ..protocols import BitrixTokenProtocol
from ._base_caller import BaseCaller
from .call_batch import call_batch


class _BatchesCaller(BaseCaller):
    """"""

    _API_METHOD: Final[Text] = "batch"
    _BATCH_RESULT_FIELDS: Final[Tuple] = ("result", "result_error", "result_total", "result_next", "result_time")
    _MAX_BATCH_SIZE: Final[int] = MAX_BATCH_SIZE

    __slots__ = ("_halt", "_methods")

    _methods: B24BatchMethods
    _halt: bool

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
            is_webhook: bool,
            methods: B24BatchMethods,
            halt: bool = False,
            timeout: Timeout = None,
            bitrix_token: Optional["BitrixTokenProtocol"] = None,
            **kwargs,
    ):
        super().__init__(
            domain=domain,
            auth_token=auth_token,
            is_webhook=is_webhook,
            api_method=self._API_METHOD,
            timeout=timeout,
            bitrix_token=bitrix_token,
            **kwargs,
        )
        self._methods = methods
        self._halt = halt

    def _fetch_batch_response(self, methods: B24BatchMethods) -> JSONDict:
        """"""
        return call_batch(
            domain=self._domain,
            auth_token=self._auth_token,
            is_webhook=self._is_webhook,
            methods=methods,
            halt=self._halt,
            timeout=self._timeout,
            **self._kwargs,
        )

    def _get_flat_methods(self) -> List[Tuple[Key, B24BatchMethodTuple]]:
        """"""
        if isinstance(self._methods, Mapping):
            return list(self._methods.items())
        else:
            return list(enumerate(self._methods))

    @staticmethod
    def _force_dict(collection: Union[Dict, List]) -> JSONDict:
        """
        Batch method can return its results in the form of either a dictionary or a list.
        This function converts results to dictionaries for uniformity.
        """
        if isinstance(collection, dict):
            return collection
        else:
            return {str(index): element for index, element in enumerate(collection)}

    def _combine_responses(self, responses: JSONList) -> JSONDict:
        """"""

        first_response, last_response = responses[0], responses[-1]

        combined_response: JSONDict = dict(
            result=dict(
                result=dict(),
                result_error=dict(),
                result_total=dict(),
                result_next=dict(),
                result_time=dict(),
            ),
            time=dict(
                start=first_response["time"]["start"],
                finish=last_response["time"]["finish"],
                duration=0,
                processing=0,
                date_start=first_response["time"]["date_start"],
                date_finish=last_response["time"]["date_finish"],
            ),
        )

        if last_response["time"].get("operating_reset_at") is not None:
            combined_response["time"]["operating_reset_at"] = last_response["time"]["operating_reset_at"]

        for response in responses:
            result = response["result"]
            time = response["time"]

            for key in self._BATCH_RESULT_FIELDS:
                combined_response["result"][key].update(self._force_dict(result.get(key, {})))

            combined_response["time"]["duration"] += time["duration"]
            combined_response["time"]["processing"] += time["processing"]

            if time.get("operating") is not None:
                combined_response["time"]["operating"] = combined_response["time"].get("operating", 0) + time["operating"]

        return combined_response

    def call(self) -> JSONDict:
        """"""

        total_methods = len(self._methods)

        if total_methods <= self._MAX_BATCH_SIZE:
            return self._fetch_batch_response(methods=self._methods)

        flat_methods: List[Tuple[Key, B24BatchMethodTuple]] = self._get_flat_methods()

        batch_responses: JSONList = list()

        for index in range(0, total_methods, self._MAX_BATCH_SIZE):
            methods_chunk = dict(flat_methods[index:index + self._MAX_BATCH_SIZE])
            batch_response = self._fetch_batch_response(methods=methods_chunk)
            batch_responses.append(batch_response)

            if self._halt and batch_response["result"]["result_error"]:
                break

        return self._combine_responses(batch_responses)


@overload
def call_batches(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        methods: Mapping[Key, B24BatchMethodTuple],
        halt: bool = False,
        timeout: Timeout = None,
        **kwargs,
) -> JSONDict: ...


@overload
def call_batches(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        methods: Sequence[B24BatchMethodTuple],
        halt: bool = False,
        timeout: Timeout = None,
        **kwargs,
) -> JSONDict: ...


def call_batches(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        methods: B24BatchMethods,
        halt: bool = False,
        timeout: Timeout = None,
        bitrix_token: Optional["BitrixTokenProtocol"] = None,
        **kwargs,
) -> JSONDict:
    """
    Using 'batch' API method, call multiple API methods in one hit to Bitrix for performance benefits. Unlike call_batch(), works with any number of API methods

    Note: one call to batch method allows you to make up to 50 actual REST API requests in one hit, mitigating requests intensity limits.

    Args:
        domain: bitrix portal domain
        auth_token: auth token
        is_webhook: whether the method is being called using webhook token
        methods:
                Collection of methods to call. Each item in a collection should be a tuple containing rest api method and its parameters.
                If the collection provided is a mapping, its keys are used to assosiate methods with their respective results.
        halt: whether to halt the sequence of requests in case of an error
        timeout: timeout in seconds
        bitrix_token:

    Returns:
        dictionary containing the result of the batch method call and information about call time
    """
    return _BatchesCaller(
        domain=domain,
        auth_token=auth_token,
        is_webhook=is_webhook,
        methods=methods,
        halt=halt,
        timeout=timeout,
        bitrix_token=bitrix_token,
        **kwargs,
    ).call()



from functools import cached_property

from .._base_scope import BaseScope
from .api import API

__all__ = [
    "Socialnetwork",
]


class Socialnetwork(BaseScope):
    """"""

    @cached_property
    def api(self) -> API:
        """"""
        return API(self)



from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Orderentity",
]


class Orderentity(BaseCRM):
    """These methods offer capabilities for managing online store orders.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/order-entity/index.html
    """

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add order binding to CRM object.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/order-entity/crm-order-entity-add.html

        This method adds a binding of an order to a CRM object.

        Args:
            fields: Object format:
                {
                    "orderID": Order identifier,

                    "ownerTypeId":  Identifier of the CRM object type,

                    "ownerId": Identifier of the CRM object
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of order bindings to CRM objects.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/order-entity/crm-order-entity-list.html

        This method returns a list of order bindings to CRM entities.

        Args:
            select: An array of fields to select;

            filter: Object format:

                {
                    "field_1": "value_1",

                    "field_2": "value_2",

                    ...,

                    "field_n": "value_n",
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def delete_by_filter(
            self,
            fields: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Remove order bindings to CRM object.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/order-entity/crm-order-entity-delete-by-filter.html

        This method removes the binding of an order to a CRM object.

        Args:
            fields: Object format:
                {
                    "orderID": Order identifier,

                    "ownerTypeId":  Identifier of the CRM object type,

                    "ownerId": Identifier of the CRM object
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete_by_filter,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get_fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get order binding fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/order-entity/crm-order-entity-get-fields.html

        This method returns a list of available order binding fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.get_fields,
            timeout=timeout,
        )



from .bitrix_app import AbstractBitrixApp, AbstractBitrixAppLocal, BitrixApp, BitrixAppLocal
from .bitrix_token import AbstractBitrixToken, AbstractBitrixTokenLocal, BitrixToken, BitrixTokenLocal, BitrixWebhook
from .oauth_placememt_data import OAuthPlacementData
from .oauth_token import OAuthToken
from .renewed_oauth_token import RenewedOAuthToken

__all__ = [
    "AbstractBitrixApp",
    "AbstractBitrixAppLocal",
    "AbstractBitrixToken",
    "AbstractBitrixTokenLocal",
    "BitrixApp",
    "BitrixAppLocal",
    "BitrixToken",
    "BitrixTokenLocal",
    "BitrixWebhook",
    "OAuthPlacementData",
    "OAuthToken",
    "RenewedOAuthToken",
]



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .base_relationship import BaseRelationship

__all__ = [
    "Company",
]


class Company(BaseRelationship):
    """The methods provide capabilities for working with companies linked to the contacts.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/contacts/company/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields for Contact-Company.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/contacts/company/crm-contact-company-fields.html

        The method returns the description of fields for the contact-company relationship.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().fields(timeout=timeout)

    @type_checker
    def add(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a Company to the specified Contact.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/contacts/company/crm-contact-company-add.html

        The method adds a company to the specified contact.

        Args:
            bitrix_id: Identifier of the contact;

            fields: Object format:
                {
                    field_1: value_1,
                    field_2: value_2,
                    ...,
                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().add(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            fields: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/contacts/company/crm-contact-company-delete.html

        This method removes a company from the specified contact.

        Args:
            bitrix_id: Identifier of the specified contact.

            fields: Object format:
                {
                    "COMPANY_ID": value
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().delete(
            bitrix_id,
            fields=fields,
            timeout=timeout,
        )



from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import Timeout
from .._base_scope import BaseScope

__all__ = [
    "App",
]


class App(BaseScope):
    """"""

    @type_checker
    def info(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.info,
            timeout=timeout,
        )



from typing import IO, Dict, Final, Optional, Text, Tuple

import requests

from ...error import BitrixRequestError, BitrixRequestTimeout
from ...utils.types import JSONDict, Number, Timeout
from ._base_requester import BaseRequester


class BitrixAPIRequester(BaseRequester):
    """"""

    _ALLOW_REDIRECTS: Final[bool] = False
    _HEADERS: Final[Dict] = {"Content-Type": "application/json"}

    __slots__ = ("_files", "_params", "_url")

    _files: Optional[Dict[Text, Tuple[Text, IO]]]
    _params: Optional[JSONDict]
    _url: Text

    def __init__(
            self,
            url: Text,
            *,
            params: Optional[JSONDict] = None,
            files: Optional[Dict[Text, Tuple[Text, IO]]] = None,
            timeout: Timeout = None,
            max_retries: Optional[int] = None,
            initial_retry_delay: Optional[Number] = None,
            retry_delay_increment: Optional[Number] = None,
    ):
        super().__init__(
            timeout=timeout,
            max_retries=max_retries,
            initial_retry_delay=initial_retry_delay,
            retry_delay_increment=retry_delay_increment,
        )
        self._url = url
        self._params = params
        self._files = files

    @property
    def _headers(self) -> Dict:
        """"""
        return self._get_default_headers() | self._HEADERS

    def _request(self) -> requests.Response:
        return requests.post(
            url=self._url,
            json=self._params,
            timeout=self._timeout,
            files=self._files,
            allow_redirects=self._ALLOW_REDIRECTS,
            headers=self._headers,
        )

    def _post(self) -> requests.Response:
        """
        Makes a POST-requests to given url

        Returns:
            Response returned by the server

        Raises:
            ConnectionToBitrixError: if failed to establish HTTP connection

            BitrixRequestTimeout: if the requests timed out
        """

        try:
            return self._request_with_retries()

        except requests.Timeout as error:
            raise BitrixRequestTimeout(timeout=self._timeout, original_error=error) from error

        except requests.RequestException as error:
            raise BitrixRequestError(original_error=error) from error

    def call(self) -> JSONDict:
        """"""
        return self._parse_response(self._post())



from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import type_checker
from ..utils.types import Timeout
from ._base_scope import BaseScope

__all__ = [
    "Server",
]


class Server(BaseScope):
    """"""

    @type_checker
    def time(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.time,
            timeout=timeout,
        )



from typing import TYPE_CHECKING, Dict, Optional, Text, Tuple

from ...utils.types import B24APIResult, JSONDict, Timeout
from ..responses import BitrixAPIResponse, BitrixAPITimeResponse

if TYPE_CHECKING:
    from ..credentials import AbstractBitrixToken
    from .bitrix_api_list_request import BitrixAPIListFastRequest, BitrixAPIListRequest


class BitrixAPIRequest:
    """"""

    __slots__ = ("_api_method", "_bitrix_token", "_kwargs", "_params", "_response", "_timeout")

    _bitrix_token: "AbstractBitrixToken"
    _api_method: Text
    _params: Optional[JSONDict]
    _timeout: Timeout
    _response: Optional[BitrixAPIResponse]
    _kwargs: Dict

    def __init__(
            self,
            *,
            bitrix_token: "AbstractBitrixToken",
            api_method: Text,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
            **kwargs: Dict,
    ):
        self._bitrix_token = bitrix_token
        self._api_method = api_method
        self._params = params
        self._timeout = timeout
        self._response = None
        self._kwargs = kwargs

    def __str__(self):
        return f"<{self.__class__.__name__} {self._api_method}({self._param_string})>"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"bitrix_token={self._bitrix_token}, "
            f"api_method='{self._api_method}', "
            f"params={self._params}, "
            f"timeout={self._timeout})"
        )

    @property
    def _as_tuple(self) -> Tuple[Text, Optional[JSONDict]]:
        """"""
        return self._api_method, self._params

    @property
    def bitrix_token(self) -> "AbstractBitrixToken":
        """"""
        return self._bitrix_token

    @property
    def api_method(self) -> Text:
        """"""
        return self._api_method

    @property
    def params(self) -> Optional[JSONDict]:
        """"""
        return self._params

    @property
    def _param_string(self) -> Text:
        """"""
        if isinstance(self._params, dict):
            return ", ".join(f"{key}={value}" for key, value in self._params.items())
        else:
            return ""

    @property
    def timeout(self) -> Timeout:
        """"""
        return self._timeout

    @property
    def response(self) -> BitrixAPIResponse:
        """"""
        return self._response or self.call()

    @property
    def result(self) -> B24APIResult:
        """"""
        return self.response.result

    @property
    def time(self) -> BitrixAPITimeResponse:
        """"""
        return self.response.time

    def _call(self) -> JSONDict:
        """"""
        return self._bitrix_token.call_method(
            api_method=self._api_method,
            params=self._params,
            timeout=self._timeout,
            **self._kwargs,
        )

    def call(self) -> BitrixAPIResponse:
        """"""
        self._response = BitrixAPIResponse.from_dict(self._call())
        return self._response

    def as_list(
            self,
            limit: Optional[int] = None,
    ) -> "BitrixAPIListRequest":
        """"""
        from .bitrix_api_list_request import BitrixAPIListRequest
        return BitrixAPIListRequest(
            bitrix_api_request=self,
            limit=limit,
            **self._kwargs,
        )

    def as_list_fast(
            self,
            descending: bool = False,
            limit: Optional[int] = None,
    ) -> "BitrixAPIListFastRequest":
        """"""
        from .bitrix_api_list_request import BitrixAPIListFastRequest
        return BitrixAPIListFastRequest(
            bitrix_api_request=self,
            descending=descending,
            limit=limit,
            **self._kwargs,
        )



from typing import Optional

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Numerator",
]


class Numerator(BaseCRM):
    """The methods provide capabilities for managing numberers.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/numerator/index.html
    """

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a new numerator.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/numerator/crm-document-generator-numerator-add.html

        The method adds a new numerator.

        Args:
            fields: Object in the format:

                {
                    name: value,

                    template: value,

                    settings: value,
                };

            timeout: Timeout in seconds;

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields=fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get information about the numerator by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/numerator/crm-document-generator-numerator-get.html

        The method returns information about the numerator by its identifier.

        Args:
            bitrix_id: Identifier of the numerator;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id=bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of numerators.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/numerator/crm-document-generator-numerator-list.html

        The method returns a list of numerators.

        Args:
            start: The parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(start=start, timeout=timeout)

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update existing numbering.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/numerator/crm-document-generator-numerator-update.html

        The method updates an existing numbering with new values.

        Args:
            bitrix_id: Identifier of the numbering;

            fields: Object in the format:

                {
                    name: value,

                    template: value,

                    settings: value,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id=bitrix_id,
            fields=fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete numerator.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/numerator/crm-document-generator-numerator-delete.html

        The method removes a numerator.

        Args:
            bitrix_id: Identifier of the numerator;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id=bitrix_id, timeout=timeout)



from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Dict, Optional

from ..._constants import PYTHON_VERSION as _PV
from ...utils.types import JSONDict

_DATACLASS_KWARGS = {"eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class BitrixAPITimeResponse:
    """"""

    start: float
    finish: float
    duration: float
    processing: float
    date_start: datetime
    date_finish: datetime
    operating_reset_at: Optional[datetime] = None
    operating: Optional[float] = None

    @classmethod
    def from_dict(cls, json_response: JSONDict) -> "BitrixAPITimeResponse":
        return cls(
            start=json_response["start"],
            finish=json_response["finish"],
            duration=json_response["duration"],
            processing=json_response["processing"],
            date_start=datetime.fromisoformat(json_response["date_start"]),
            date_finish=datetime.fromisoformat(json_response["date_finish"]),
            operating_reset_at=json_response.get("operating_reset_at") and datetime.fromtimestamp(json_response["operating_reset_at"]).astimezone(),
            operating=json_response.get("operating"),
        )

    def to_dict(self) -> Dict:
        return asdict(self)



from functools import cached_property
from typing import Iterable, Literal, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM
from .volatile_type import VolatileType

__all__ = [
    "Duplicate",
]


class Duplicate(BaseCRM):
    """Method for finding duplicates in leads, contacts or companies depending on emails or phone numbers.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/index.html
    """

    @cached_property
    def volatile_type(self) -> VolatileType:
        """"""
        return VolatileType(self)

    @type_checker
    def find_by_comm(
            self,
            *,
            type: Literal["EMAIL", "PHONE"],
            values: Iterable[Text],
            entity_type: Optional[Literal["LEAD", "CONTACT", "COMPANY"]] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get leads, contacts, and companies with matching data

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/crm-duplicate-find-by-comm.html

        The method returns the identifiers of lead, contacts and companies that contain phone numbers or email addresses from a specified list.

        Args:
            type: Type of communication, where possible values:

                - EMAIL,

                - PHONE;

            values: Array of emails or phone numbers (maximum number of values - 20);

            entity_type: Type of object, where possible values are:

                - LEAD,

                - CONTACT,

                - COMPANY;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest

        """

        if values.__class__ is not list:
            values = list(values)

        params = {
            "type": type,
            "values": values,
        }

        if entity_type is not None:
            params["entity_type"] = entity_type

        return self._make_bitrix_api_request(
            api_wrapper=self.find_by_comm,
            params=params,
            timeout=timeout,
        )



import datetime as dt

import pytest

from b24pysdk.bitrix_api.responses import BitrixAPITimeResponse


def test_response_time_from_dict_and_to_dict():
    now = dt.datetime.now(dt.timezone.utc)
    later = now + dt.timedelta(seconds=1)

    payload = {
        "start": 0.1,
        "finish": 0.2,
        "duration": 0.1,
        "processing": 0.05,
        "date_start": now.isoformat(),
        "date_finish": later.isoformat(),
        "operating_reset_at": int(now.timestamp()),
        "operating": 0.08,
    }

    rt = BitrixAPITimeResponse.from_dict(payload)
    assert rt.duration == pytest.approx(0.1)
    assert rt.processing == pytest.approx(0.05)
    assert isinstance(rt.date_start, dt.datetime)
    assert isinstance(rt.date_finish, dt.datetime)

    # Round-trip via to_dict
    d = rt.to_dict()
    # dataasdict converts datetimes to isoformat? It preserves objects; check fields presence
    assert set(d.keys()) >= {"start", "finish", "duration", "processing", "date_start", "date_finish"}



from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Multifield",
]


class Multifield(BaseCRM):
    """"""

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of multiple fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/multifield/crm-multifield-fields.html

        The method returns the description of multiple fields used to store phone numbers, email addresses, and other contact information in leads, contacts and companies.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)



from typing import Optional

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Logmessage",
]


class Logmessage(BaseCRM):
    """These methods offer capabilities for working with the log message journal, where the log message journal is a special type of timeline record.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/index.html
    """

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add log entry.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/crm-timeline-logmessage-add.html

        This method adds a new log entry to the timeline.

        Args:
            fields: Object format:

                {
                    entityTypeId: "value",

                    entityId: "value",

                    title: "value",

                    text: "value",

                    iconCode: "value",
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get information about the log entry.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/crm-timeline-logmessage-get.html

        This method retrieves information about a timeline log entry.

        Args:
            bitrix_id: Integer identifier of the timeline entry;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            entity_type_id: int,
            entity_id: id,
            *,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of log entries.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/crm-timeline-logmessage-list.html

        This method retrieves a list of timeline log entries.

        Args:
            entity_type_id: Identifier of the entity type for which to retrieve the list of log entries;

            entity_id: Identifier of the entity item for which to retrieve the list of log entries;

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        if order is not None:
            params["order"] = order

        if start is not None:
            params["start"] = start

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete log entry.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/logmessage/crm-timeline-logmessage-delete.html

        This method deletes a timeline log entry.

        Args:
            bitrix_id: Integer identifier of the timeline entry;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(
            bitrix_id,
            timeout=timeout,
        )



from abc import abstractmethod
from functools import cached_property

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM
from .items import Items


class BaseRelationship(BaseCRM):
    """These methods provide functionality for working with Company-Contact and
    Contact-CRM entity bindings, where the CRM entity can be a Lead, Deal, or Company."""

    @cached_property
    def items(self) -> "Items":
        """"""
        return Items(self)

    @abstractmethod
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields for the connection between entities.

        This method retrieves the fields of the connection between a Contact or Company and the specified CRM entity (Lead, Deal, Contact, or Company).
        Note that a Contact can be linked to a Lead, Deal, or Company, and a Company can be linked to a Contact.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @abstractmethod
    def add(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a connection between CRM entities.

        This method adds connection between Contact or Company and the specified CRM entity (Lead, Deal, Contact, or Company).
        Note that a Contact can be linked to a Lead, Deal, or Company, and a Company can be linked to a Contact.

        Args:
            bitrix_id: Identifier of the CRM entity to be linked;

            fields: Object format:
                {
                    "CONTACT_ID" or "COMPANY_ID": "value",

                    "SORT": "value",

                    "IS_PRIMARY": "value"
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @abstractmethod
    def delete(
            self,
            bitrix_id: int,
            *,
            fields: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete connection between CRM entities.

        The method removes a Contact from a Lead, Deal, or Company, and removes a Company from a Contact.

        Args:
            bitrix_id: Identifier of the CRM entity to be linked;

            fields: Object format:
                {
                    "COMPANY_ID" or "CONTACT_ID": value
                }

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Entity",
]


class Entity(BaseCRM):
    """Method lets you to merge duplicates in CRM.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/index.html
    """

    @type_checker
    def merge_batch(
            self,
            *,
            params: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Merge duplicates

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/duplicates/crm-entity-merge-batch.html

        The method merges multiple entities into one.

        Args:
            params: Object containing the entities to merge, which contains:

                - entityTypeId: Identifier of the CRM object type,

                - entityIds: array of identifiers of the entities to be merged;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "params": params,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.merge_batch,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Literal, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM
from .items import Items

__all__ = [
    "Calllist",
]


class Calllist(BaseCRM):
    """The class provides capabilities for managing methods that allows to call multiple clients in succession.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/index.html
    """

    @cached_property
    def items(self) -> Items:
        """"""
        return Items(self)

    @type_checker
    def add(
            self,
            *,
            entity_type: Literal["CONTACT", "COMPANY"],
            entities: Iterable[int],
            webform_id: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new call list.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/crm-calllist-add.html

        The method creates a new call list.

        Args:
            entity_type: Type of the object:

                - CONTACT - for contact,

                - COMPANY - for company;

            entities: Array of IDs of contacts or companies;

            webform_id: ID of the CRM form that will be displayed in call list form;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        if entities.__class__ is not list:
            entities = list(entities)

        params = {
            "ENTITY_TYPE": entity_type,
            "ENTITIES": entities,
        }

        if webform_id is not None:
            params["WEBFORM_ID"] = webform_id

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get information about the call list.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/crm-calllist-get.html

        The method returns information about the call list by its identifier, without the list of participants.

        Args:
            bitrix_id: Identifier of the call list;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ID": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of call lists.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/crm-calllist-list.html

        The method returns a list of call activities.

        Args:
            select: Array of fields to retrieve;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        params = dict()

        if select is not None:
            if select.__class__ is not list:
                select = list(select)

            params["SELECT"] = select

        if filter is not None:
            params["FILTER"] = filter

        if order is not None:
            params["ORDER"] = order

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def statuslist(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of call statuses.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/crm-calllist-statuslist.html

        The method returns a list of call statuses.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.statuslist,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            *,
            list_id: int,
            entity_type: Literal["CONTACT", "COMPANY"],
            entities: Iterable[int],
            webform_id: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update call list composition.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/call-list/crm-calllist-update.html

        The method allows you to add or remove participants from an existing call list and update the associated CRM form.

        Args:
            list_id: Identifier of the call list;

            entity_type: Type of the object:

                - CONTACT - for contact,

                - COMPANY - for company;

            entities: Array of IDs of contacts or companies;

            webform_id: ID of the CRM form that will be displayed in call list form;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        if entities.__class__ is not list:
            entities = list(entities)

        params = {
            "LIST_ID": list_id,
            "ENTITY_TYPE": entity_type,
            "ENTITIES": entities,
        }

        if webform_id is not None:
            params["WEBFORM_ID"] = webform_id

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )



from abc import ABC

from ._base_context import BaseContext


class BaseEntity(BaseContext, ABC):
    """"""

    __slots__ = ("_context",)

    _context: BaseContext

    def __init__(self, context: BaseContext):
        self._context = context

    def __repr__(self):
        return f"client.{self}"



from dataclasses import dataclass

from ..._constants import PYTHON_VERSION as _PV

_DATACLASS_KWARGS = {"eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class BaseBitrixEvent:
    """"""



from typing import Optional

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from .._base_entity import BaseEntity

__all__ = [
    "Userfield",
]


class Userfield(BaseEntity):
    """Manage user-defined fields in the Bitrix24 system.

    Documentation: https://apidocs.bitrix24.com/api-reference/user/userfields/index.html
    """

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Add a custom field.

        Documentation: https://apidocs.bitrix24.com/api-reference/user/userfields/user-userfield-add.html

        This method allows adding custom fields to the existing entities.

        Args:
            fields: A dictionary of field parameters to create;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            order: Optional[JSONDict] = None,
            filter: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Retrieve a list of custom fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/user/userfields/user-userfield-list.html

        This method retrieves a list of user fields with optional ordering and filtering.

        Args:
            order: A dictionary to sort the fields;

            filter: A dictionary to filter the fields;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = dict()

        if order is not None:
            params["order"] = order

        if filter is not None:
            params["filter"] = filter

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Update a custom field.

        Documentation: https://apidocs.bitrix24.com/api-reference/user/userfields/user-userfield-update.html

        This method modifies the properties of the specified user field.

        Args:
            bitrix_id: The ID of the user field to update;

            fields: A dictionary of field parameters to update;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Delete a custom field

        Documentation: https://apidocs.bitrix24.com/api-reference/user/userfields/user-userfield-delete.html

        This method removes a custom field from the system.

        Args:
            bitrix_id: The ID of the user field to delete;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Type",
]


class Type(BaseCRM):
    """The methods provide capabilities for working with custom activity types.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/types/index.html
    """

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add custom CRM activity type.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/types/crm-activity-type-add.html

        The method register a custom activity type by specifying its name and icon.

        Args:
            fields: Object format:

                {
                    "TYPE_ID": 'value',

                    "NAME": 'value',

                    "ICON_FILE": 'value',

                    "IS_CONFIGURABLE_TYPE": 'value',
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of custom activity types.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/types/crm-activity-type-list.html

        The method retrieves a list of custom activity types registered by the application.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(timeout=timeout)

    @type_checker
    def delete(
            self,
            type_id: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete custom activity type.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/types/crm-activity-type-delete.html

        The method removes a custom activity type.

        Args:
            type_id: String value of the activity type;

            timeout: Timeout in seconds.

        Returns:
            BitrixAPIRequest
        """

        params = {
            "TYPE_ID": type_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Document",
]


class Document(BaseCRM):
    """The methods provide capabilities for working with documents.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/index.html
    """

    @type_checker
    def getfields(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get document fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-get-fields.html

        The method returns a list of document fields along with their description.

        Args:
            bitrix_id: Document identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.getfields,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            *,
            template_id: int,
            entity_type_id: int,
            entity_id: int,
            values: JSONDict,
            stamps_enabled: bool,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new document.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-add.html

        The method creates a new document based on a template and data from the corresponding entity.

        Args:
            template_id: Identifier of the template;

            entity_type_id: CRM entity type ID;

            entity_id: Identifier of the CRM entity;

            values: Additional field values;

            stamps_enabled: Stamps and signatures, where True - enable, False - disable;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "templateId": template_id,
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "values": values,
            "stampsEnabled": stamps_enabled,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get document information.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-get.html

        The method returns information about a document by its identifier.

        Args:
            bitrix_id: Identifier of the document;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id=bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            order: Optional[JSONDict] = None,
            filter: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of documents.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-list.html

        The method returns a list of documents based on the filter.

        Args:
            select: Array of fields to output;

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            order=order,
            filter=filter,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            *,
            values: JSONDict,
            stamps_enabled: bool,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update document.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-update.html

        The method updates an existing document with new values

        Args:
            bitrix_id: Document identifier;

            values: Array of new field values for the document;

            stamps_enabled: Stamps and signatures, where True - enable, False - disable;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "values": values,
            "stampsEnabled": stamps_enabled,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete document.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-delete.html

        The method removes a document.

        Args:
            bitrix_id: Document identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id=bitrix_id, timeout=timeout)

    @type_checker
    def enablepublicurl(
            self,
            bitrix_id: int,
            *,
            status: bool,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Enable and disable public link for document.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-enable-public-url.html

        The method enables/disables thr public link for the document.

        Args:
            bitrix_id: Document identifier;

            status: Status of the public link for the document, where True - enable, False - disable;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "status": status,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.enablepublicurl,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def upload(
            self,
            *,
            file_content: Text,
            region: Text,
            entity_type_id: int,
            entity_id: int,
            title: Text,
            number: int,
            pdf_content: Optional[Text] = None,
            image_content: Optional[Text] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Upload and attach the generated document.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/document-generator/documents/crm-document-generator-document-upload.html

        The method uploads a generated document and attaches it to the specified entity.

        Args:
            file_content: Content of the docx file in base64;

            region: Country;

            entity_type_id: Identifier of the CRM entity type;

            entity_id: Identifier of the CRM entity;

            title: Document title;

            number: Document number;

            pdf_content: Content of the pdf file in base64;

            image_content: Content of the image in base64;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fileContent": file_content,
            "region": region,
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "title": title,
            "number": number,
        }

        if pdf_content is not None:
            params["pdfContent"] = pdf_content

        if image_content is not None:
            params["imageContent"] = image_content

        return self._make_bitrix_api_request(
            api_wrapper=self.upload,
            params=params,
            timeout=timeout,
        )



from typing import Optional, Text

from ...utils.types import JSONDict, Timeout
from ..protocols import BitrixTokenProtocol
from ._base_caller import BaseCaller
from .call import call


class _MethodCaller(BaseCaller):
    """"""

    def __init__(
            self,
            domain: Text,
            auth_token: Text,
            is_webhook: bool,
            api_method: Text,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
            bitrix_token: Optional["BitrixTokenProtocol"] = None,
            **kwargs,
    ):
        super().__init__(
            domain=domain,
            auth_token=auth_token,
            is_webhook=is_webhook,
            api_method=api_method,
            params=params,
            timeout=timeout,
            bitrix_token=bitrix_token,
            **kwargs,
        )

    @property
    def _dynamic_auth_token(self) -> Text:
        """"""
        return ("", f"{self._auth_token}/")[self._is_webhook]

    @property
    def _url(self) -> Text:
        """"""
        return f"https://{self._domain}/rest/{self._dynamic_auth_token}{self._api_method}.json"

    @property
    def _dynamic_params(self) -> JSONDict:
        """"""
        if self._is_webhook:
            return self._params
        else:
            return self._params | {"auth": self._auth_token}

    def call(self) -> JSONDict:
        """"""
        self._config.logger.debug(
            "start call_method",
            context=dict(
                domain=self._domain,
                method=self._api_method,
                parameters=self._params,
            ),
        )
        json_response = call(
                url=self._url,
                params=self._dynamic_params,
                timeout=self._timeout,
                **self._kwargs,
        )
        self._config.logger.debug(
            "finish call_method",
            context=dict(
                result=json_response["result"],
                time=json_response["time"],
            ),
        )
        return json_response


def call_method(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        api_method: Text,
        params: Optional[JSONDict] = None,
        timeout: Timeout = None,
        bitrix_token: Optional["BitrixTokenProtocol"] = None,
        **kwargs,
) -> JSONDict:
    """
    Call a Bitrix API method

    Args:
        domain: bitrix portal domain
        auth_token: auth token
        is_webhook: whether the method is being called using webhook token
        api_method: name of the bitrix API method to call, e.g. crm.deal.add
        params: API method parameters
        timeout: timeout in seconds
        bitrix_token:

    Returns:
        dictionary containing the result of the API method call and information about call time
    """
    return _MethodCaller(
        domain=domain,
        auth_token=auth_token,
        is_webhook=is_webhook,
        api_method=api_method,
        params=params,
        timeout=timeout,
        bitrix_token=bitrix_token,
        **kwargs,
    ).call()



from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import B24Bool, JSONDict, Timeout
from .._base_entity import BaseEntity

__all__ = [
    "Workgroup",
]


class Workgroup(BaseEntity):
    """Handle operations related to Bitrix24 workgroups.

    Documentation: https://apidocs.bitrix24.com/api-reference/sonet-group/index.html
    """

    @type_checker
    def get(
            self,
            params: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """
        Retrieve information about a specific workgroup.

        Documentation: https://apidocs.bitrix24.com/api-reference/sonet-group/socialnetwork-api-workgroup-get.html

        This method fetches details of the workgroup based on the provided parameters.

        Args:
            params: Parameters to be sent with the API requests;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest.
        """

        params = {
            "params": params,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            filter: Optional[JSONDict] = None,
            select: Optional[Iterable[Text]] = None,
            is_admin: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Retrieve a list of workgroups based on provided filters.

        Documentation: https://apidocs.bitrix24.com/api-reference/sonet-group/socialnetwork-api-workgroup-list.html

        This method fetches a list of workgroups that match the filter criteria and selection fields.

        Args:
            filter: A dictionary containing filters to apply;

            select: An iterable of field names to include in the responses;

            is_admin: Boolean to specify if admin-specific data is required;

            timeout: Timeout setting for the requests in seconds.

        Returns:
            Instance of BitrixAPIRequest.
        """

        params = dict()

        if filter is not None:
            params["filter"] = filter

        if select is not None:
            if select.__class__ is not list:
                select = list(select)

            params["select"] = select

        if is_admin is not None:
            params["IS_ADMIN"] = B24Bool(is_admin).to_str()

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._productrows import Productrows
from ._relationships import Contact
from ._userfield import Userfield
from .details import Details
from .item.base_item import BaseItem

__all__ = [
    "Lead",
]


class Lead(BaseItem):
    """The methods provide capabilities for managing leads.
    They allow you to retrieve fields, add, update, delete, and get lists of leads.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/index.html
    """

    @cached_property
    def contact(self) -> Contact:
        """"""
        return Contact(self)

    @cached_property
    def details(self) -> Details:
        """"""
        return Details(self)

    @cached_property
    def productrows(self) -> Productrows:
        """"""
        return Productrows(self)

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get CRM lead fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/crm-lead-fields.html

        The method returns the description of lead fields, including custom fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new lead.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/crm-lead-add.html

        The method creates a new lead.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            params: Set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register additional event 'Y' or not 'N',

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        return self._add(
            fields,
            extra_params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get lead by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/crm-lead-get.html

        The method returns a lead by its identifier.

        Args:
            bitrix_id: The identifier of the lead;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of leads.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/crm-lead-list.html

        The method returns a list of leads based on a filter.

        Args:
            select: An array containing the list of fields to be selected;

            filter: Object format:

                {
                    "field_1": "value_1",

                    "field_2": "value_2",

                    ...,

                    "field_n": "value_n",
                }

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used to control pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update lead.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/crm-lead-update.html

        The method updates an existing lead.

        Args:
            bitrix_id: lead identifier;

            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            params: Set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the activity stream 'Y' or not 'N';

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            extra_params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete lead.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/leads/crm-lead-delete.html

        The method removes a lead and all associated objects: tasks, history, timeline records, and others.

        Objects are deleted if they are not linked to other objects or entities. If the objects are linked to other entities, only the link to the deleted lead will be removed.

        Args:
            bitrix_id: Identifier of the lead;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)



from abc import abstractmethod
from typing import Iterable, Optional, Text

from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.types import JSONDict, Timeout
from ..._base_crm import BaseCRM

__all__ = [
    "BaseDetail",
]


class BaseDetail(BaseCRM):
    """The methods provide capabilities for managing requisite templates and bank details."""

    @abstractmethod
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of the entity fields.

        The method returns a formal description of the fields of the requisite template or bank details.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @abstractmethod
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new entity.

        This method creates a new requisites template or bank details.

        Args:
            fields: A set of fields - an object for adding a new entity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields=fields, timeout=timeout)

    @abstractmethod
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Get entity by ID.

        This method returns the requisite template or bank details by its identifier.

        Args:
            bitrix_id: Identifier of the entity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id=bitrix_id, timeout=timeout)

    @abstractmethod
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of entities.

        The method returns a list of requisites templates or bank details based on the filter.

        Args:
            select: An array containing the list of fields to select;

            filter: Object format:
                {
                    field_1: value_1,
                    field_2: value_2,
                    ...,
                    field_n: value_n,
                };

            order: Object format:
                {
                    field_1: value_1,
                    field_2: value_2,
                    ...,
                    field_n: value_n,
                },

                where

                    - field_n is the name of the field by which the selection will be sorted

                    - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @abstractmethod
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update entity.

        This method updates the requisite template or bank details.

        Args:
            bitrix_id: Identifier of the entity to be updated;

            fields: Set of template fields — an object, the values of which need to be changed;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id=bitrix_id,
            fields=fields,
            timeout=timeout,
        )

    @abstractmethod
    def delete(
            self,
            bitrix_id,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete entity.

        This method deletes requisite template or bank entity.

        Args:
            bitrix_id: Identifier of the entity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id=bitrix_id, timeout=timeout)



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .base_relationship import BaseRelationship

__all__ = [
    "Contact",
]


class Contact(BaseRelationship):
    """These methods provide capabilities for managing Contacts linked to the Deals, Leads and Companies (CRM entity).

    Documentation:
        - for deals: https://apidocs.bitrix24.com/api-reference/crm/deals/contacts/index.html
        - for leads: https://apidocs.bitrix24.com/api-reference/crm/leads/management-communication/index.html
        - for companies: https://apidocs.bitrix24.com/api-reference/crm/companies/contacts/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields for entity-contact connection.

        Documentation:

        - `for leads; <https://apidocs.bitrix24.com/api-reference/crm/leads/management-communication/crm-lead-contact-fields.html>`__

        - `for deals; <https://apidocs.bitrix24.com/api-reference/crm/deals/contacts/crm-deal-contact-fields.html>`__

        - `for companies. <https://apidocs.bitrix24.com/api-reference/crm/companies/contacts/crm-company-contact-fields.html>`__

        The method retrieves the description of the fields for the entity-contact relationship, where entity is one of the available CRM-entities: Lead, Deal or Company.

        Args:
             timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().fields(timeout=timeout)

    @type_checker
    def add(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add contact binding to CRM entity.

        Documentation:

        - `for leads; <https://apidocs.bitrix24.com/api-reference/crm/leads/management-communication/crm-lead-contact-add.html>`__

        - `for deals; <https://apidocs.bitrix24.com/api-reference/crm/deals/contacts/crm-deal-contact-add.html>`__

        - `for companies. <https://apidocs.bitrix24.com/api-reference/crm/companies/contacts/crm-company-contact-add.html>`__

        This method adds a contact binding to the specified CRM entity, where entity is one of the available CRM-entities: Lead, Deal or Company.

        Args:
            bitrix_id: The identifier of the CRM entity to which the contact needs to be added;

            fields: Object format:
                {
                    "CONTACT_ID": "value",

                    "SORT": "value",

                    "IS_PRIMARY": "value"
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().add(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            fields: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Remove contact from CRM entity.

        Documentation:

        - `for leads; <https://apidocs.bitrix24.com/api-reference/crm/leads/management-communication/crm-lead-contact-delete.html>`__

        - `for deals; <https://apidocs.bitrix24.com/api-reference/crm/deals/contacts/crm-deal-contact-delete.html>`__

        - `for companies. <https://apidocs.bitrix24.com/api-reference/crm/companies/contacts/crm-company-contact-delete.html>`__

        The method removes a contact from the specified CRM entity, where entity is one of the available CRM-entities: Lead, Deal or Company.

        Args:
            bitrix_id: The identifier of the CRM entity from which to remove the contact binding;

            fields: Object format:
                {
                    "CONTACT_ID": "value",
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return super().delete(
            bitrix_id,
            fields=fields,
            timeout=timeout,
        )



from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Text
from urllib.parse import urlparse

from ..._constants import PYTHON_VERSION as _PV
from ...constants import B24AppStatus
from ...error import BitrixValidationError
from ...utils.types import JSONDict
from .oauth_token import OAuthToken

_DATACLASS_KWARGS = {"eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class RenewedOAuthToken:
    """"""

    class ValidationError(BitrixValidationError):
        """"""

    oauth_token: OAuthToken
    member_id: Text
    client_endpoint: Text
    server_endpoint: Text
    domain: Text
    scope: List[Text]
    status: B24AppStatus
    application_token: Optional[Text] = None

    @classmethod
    def from_dict(cls, renewed_oauth_token_payload: JSONDict) -> "RenewedOAuthToken":
        try:
            return cls(
                oauth_token=OAuthToken.from_dict(renewed_oauth_token_payload),
                member_id=renewed_oauth_token_payload["member_id"],
                client_endpoint=renewed_oauth_token_payload["client_endpoint"],
                server_endpoint=renewed_oauth_token_payload["server_endpoint"],
                domain=renewed_oauth_token_payload["domain"],
                scope=renewed_oauth_token_payload["scope"].split(","),
                status=B24AppStatus(renewed_oauth_token_payload["status"]),
                application_token=renewed_oauth_token_payload.get("application_token"),
            )
        except KeyError as error:
            raise cls.ValidationError(f"Missing required field in renewed OAuth token payload: {error.args[0]}") from error
        except Exception as error:
            raise cls.ValidationError(f"Invalid renewed OAuth token payload: {error}") from error

    @property
    def portal_domain(self) -> Text:
        """"""
        return urlparse(self.client_endpoint).hostname

    def to_dict(self) -> Dict:
        return asdict(self)



from .base_bitrix_event import BaseBitrixEvent
from .oauth_token_renewed_event import OAuthTokenRenewedEvent
from .portal_domain_changed_event import PortalDomainChangedEvent

__all__ = [
    "BaseBitrixEvent",
    "OAuthTokenRenewedEvent",
    "PortalDomainChangedEvent",
]



from functools import cached_property

from .._base_entity import BaseEntity
from .workgroup import Workgroup

__all__ = [
    "API",
]


class API(BaseEntity):
    """"""

    @cached_property
    def workgroup(self) -> Workgroup:
        """"""
        return Workgroup(self)



from ....bitrix_api.requests import BitrixAPIRequest
from ....scopes.crm._base_crm import BaseCRM
from ....utils.functional import type_checker
from ....utils.types import Timeout

__all__ = [
    "Settings",
]


class Settings(BaseCRM):
    """"""

    @type_checker
    def mode(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of CRM operation modes.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/crm-enum-settings-mode.html

        The method returns a list of CRM operation modes.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.mode,
            timeout=timeout,
        )



from typing import Dict, Text, Type

import requests
from requests.exceptions import HTTPError, JSONDecodeError

from ....error import (
    BitrixAPIAccessDenied,
    BitrixAPIAllowedOnlyIntranetUser,
    BitrixAPIBadRequest,
    BitrixAPIError,
    BitrixAPIErrorBatchLengthExceeded,
    BitrixAPIErrorBatchMethodNotAllowed,
    BitrixAPIErrorManifestIsNotAvailable,
    BitrixAPIErrorOAuth,
    BitrixAPIErrorUnexpectedAnswer,
    BitrixAPIExpiredToken,
    BitrixAPIForbidden,
    BitrixAPIInsufficientScope,
    BitrixAPIInternalServerError,
    BitrixAPIInvalidArgValue,
    BitrixAPIInvalidCredentials,
    BitrixAPIInvalidRequest,
    BitrixAPIMethodConfirmDenied,
    BitrixAPIMethodConfirmWaiting,
    BitrixAPIMethodNotAllowed,
    BitrixAPINoAuthFound,
    BitrixAPINotFound,
    BitrixAPIOverloadLimit,
    BitrixAPIQueryLimitExceeded,
    BitrixAPIServiceUnavailable,
    BitrixAPIUnauthorized,
    BitrixAPIUserAccessError,
    BitrixOAuthInvalidClient,
    BitrixOAuthInvalidGrant,
    BitrixOAuthInvalidScope,
    BitrixOauthWrongClient,
    BitrixResponse302JSONDecodeError,
    BitrixResponse500JSONDecodeError,
    BitrixResponseJSONDecodeError,
)
from ....utils.types import JSONDict

_EXCEPTIONS_BY_ERROR: Dict[Text, Type[BitrixAPIError]] = {
    # 200
    "WRONG_CLIENT": BitrixOauthWrongClient,
    # 400
    "ERROR_BATCH_LENGTH_EXCEEDED": BitrixAPIErrorBatchLengthExceeded,
    "INVALID_ARG_VALUE": BitrixAPIInvalidArgValue,
    "INVALID_CLIENT": BitrixOAuthInvalidClient,
    "INVALID_GRANT": BitrixOAuthInvalidGrant,
    "INVALID_REQUEST": BitrixAPIInvalidRequest,
    # 401
    "ERROR_OAUTH": BitrixAPIErrorOAuth,
    "EXPIRED_TOKEN": BitrixAPIExpiredToken,
    "METHOD_CONFIRM_WAITING": BitrixAPIMethodConfirmWaiting,
    "NO_AUTH_FOUND": BitrixAPINoAuthFound,
    # 403
    "ACCESS_DENIED": BitrixAPIAccessDenied,
    "ALLOWED_ONLY_INTRANET_USER": BitrixAPIAllowedOnlyIntranetUser,
    "INSUFFICIENT_SCOPE": BitrixAPIInsufficientScope,
    "INVALID_CREDENTIALS": BitrixAPIInvalidCredentials,
    "METHOD_CONFIRM_DENIED": BitrixAPIMethodConfirmDenied,
    "USER_ACCESS_ERROR": BitrixAPIUserAccessError,
    "INVALID_SCOPE": BitrixOAuthInvalidScope,
    # 404
    "NOT_FOUND": BitrixAPINotFound,
    "ERROR_MANIFEST_IS_NOT_AVAILABLE": BitrixAPIErrorManifestIsNotAvailable,
    # 405
    "ERROR_BATCH_METHOD_NOT_ALLOWED": BitrixAPIErrorBatchMethodNotAllowed,
    # 500
    "ERROR_UNEXPECTED_ANSWER": BitrixAPIErrorUnexpectedAnswer,
    "INTERNAL_SERVER_ERROR": BitrixAPIInternalServerError,
    # 503
    "OVERLOAD_LIMIT": BitrixAPIOverloadLimit,
    "QUERY_LIMIT_EXCEEDED": BitrixAPIQueryLimitExceeded,
}
""""""

_EXCEPTIONS_BY_STATUS_CODE: Dict[int, Type[BitrixAPIError]] = {
    400: BitrixAPIBadRequest,
    401: BitrixAPIUnauthorized,
    403: BitrixAPIForbidden,
    404: BitrixAPINotFound,
    405: BitrixAPIMethodNotAllowed,
    500: BitrixAPIInternalServerError,
    503: BitrixAPIServiceUnavailable,
}
""""""

_EXCEPTIONS_BY_JSON_DECODE_RESPONSE_STATUS_CODE: Dict[int, Type[BitrixResponseJSONDecodeError]] = {
    302: BitrixResponse302JSONDecodeError,
    500: BitrixResponse500JSONDecodeError,
}


def _raise_http_error(response: requests.Response):
    raise HTTPError(
        f"{response.status_code} Client Error: {response.json()['error']} for url: {response.url}",
        response=response,
    )


def parse_response(response: requests.Response) -> JSONDict:
    """
    Parses the responses from the API server. If responses body contains an error message, raises appropriate exception

    Args:
        response: responses returned by the API server

    Returns:
        dictionary containing the parsed responses of the API server

    Raises:
        BitrixAPIError: base class for all API-related errors. Depening on an error code and/or an HTTP status code, more specific exception subclassed from BitrixAPIError will be raised.
                        These exceptions indicate that the API server successfully processed the responses, but some occured during API method execution.
        BitrixResponseJSONDecodeError: if responses returned by the API server is not a valid JSON
    """

    try:
        json_response = response.json()
    except JSONDecodeError as error:
        exception_class = (
                _EXCEPTIONS_BY_JSON_DECODE_RESPONSE_STATUS_CODE.get(response.status_code)
                or BitrixResponseJSONDecodeError
        )
        raise exception_class(original_error=error, response=response) from error

    try:
        response.raise_for_status()

        if "error" in json_response:
            _raise_http_error(response)

    except HTTPError as error:
        exception_class = (
                _EXCEPTIONS_BY_ERROR.get(str(json_response.get("error", "")).upper()) or
                _EXCEPTIONS_BY_STATUS_CODE.get(response.status_code) or
                BitrixAPIError
        )
        raise exception_class(json_response=json_response, response=response) from error

    else:
        return json_response



from abc import ABC
from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.types import JSONDict, Timeout
from .._base_entity import BaseEntity


class BaseCRM(BaseEntity, ABC):
    """"""

    def _fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self._fields,
            timeout=timeout,
        )

    def _add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self._add,
            params=params,
            timeout=timeout,
        )

    def _get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self._get,
            params=params,
            timeout=timeout,
        )

    def _list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = dict()

        if select is not None:
            if select.__class__ is not list:
                select = list(select)

            params["select"] = select

        if filter is not None:
            params["filter"] = filter

        if order is not None:
            params["order"] = order

        if start is not None:
            params["start"] = start

        return self._make_bitrix_api_request(
            api_wrapper=self._list,
            params=params,
            timeout=timeout,
        )

    def _update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "id": bitrix_id,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self._update,
            params=params,
            timeout=timeout,
        )

    def _delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self._delete,
            params=params,
            timeout=timeout,
        )



from typing import Mapping, Sequence, Union, overload

from . import scopes
from .bitrix_api.credentials import AbstractBitrixToken
from .bitrix_api.requests import BitrixAPIBatchesRequest, BitrixAPIBatchRequest, BitrixAPIRequest
from .utils.types import JSONDict, Key, Timeout


class Client:
    """"""

    __slots__ = (
        "_bitrix_token",
        "_kwargs",
        "access",
        "app",
        "crm",
        "department",
        "feature",
        "method",
        "placement",
        "profile",
        "scope",
        "server",
        "socialnetwork",
        "user",
    )

    _bitrix_token: AbstractBitrixToken
    _kwargs: JSONDict
    access: scopes.Access
    app: scopes.App
    crm: scopes.CRM
    department: scopes.Department
    feature: scopes.Feature
    method: scopes.Method
    placement: scopes.Placement
    profile: scopes.Profile
    scope: scopes.Scope
    server: scopes.Server
    socialnetwork: scopes.Socialnetwork
    user: scopes.User

    def __init__(
            self,
            bitrix_token: AbstractBitrixToken,
            **kwargs,
    ):
        self._bitrix_token = bitrix_token
        self._kwargs = kwargs
        self.access = scopes.Access(self)
        self.app = scopes.App(self)
        self.crm = scopes.CRM(self)
        self.department = scopes.Department(self)
        self.feature = scopes.Feature(self)
        self.method = scopes.Method(self)
        self.placement = scopes.Placement(self)
        self.profile = scopes.Profile(self)
        self.scope = scopes.Scope(self)
        self.server = scopes.Server(self)
        self.socialnetwork = scopes.Socialnetwork(self)
        self.user = scopes.User(self)

    def __str__(self):
        return f"<Client of portal {self._bitrix_token.domain}>"

    def __repr__(self):
        return f"{self.__class__.__name__}(bitrix_token={self._bitrix_token})"

    @overload
    def call_batch(
            self,
            bitrix_api_requests: Mapping[Key, BitrixAPIRequest],
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
    ) -> BitrixAPIBatchRequest: ...

    @overload
    def call_batch(
            self,
            bitrix_api_requests: Sequence[BitrixAPIRequest],
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
    ) -> BitrixAPIBatchRequest: ...

    def call_batch(
            self,
            bitrix_api_requests: Union[Mapping[Key, BitrixAPIRequest], Sequence[BitrixAPIRequest]],
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
    ) -> BitrixAPIBatchRequest:
        """"""
        return BitrixAPIBatchRequest(
            bitrix_token=self._bitrix_token,
            bitrix_api_requests=bitrix_api_requests,
            halt=halt,
            ignore_size_limit=ignore_size_limit,
            timeout=timeout,
            **self._kwargs,
        )

    @overload
    def call_batches(
            self,
            bitrix_api_requests: Mapping[Key, BitrixAPIRequest],
            halt: bool = False,
            timeout: Timeout = None,
    ) -> BitrixAPIBatchRequest: ...

    @overload
    def call_batches(
            self,
            bitrix_api_requests: Sequence[BitrixAPIRequest],
            halt: bool = False,
            timeout: Timeout = None,
    ) -> BitrixAPIBatchRequest: ...

    def call_batches(
            self,
            bitrix_api_requests: Union[Mapping[Key, BitrixAPIRequest], Sequence[BitrixAPIRequest]],
            halt: bool = False,
            timeout: Timeout = None,
    ) -> BitrixAPIBatchesRequest:
        """"""
        return BitrixAPIBatchesRequest(
            bitrix_token=self._bitrix_token,
            bitrix_api_requests=bitrix_api_requests,
            halt=halt,
            timeout=timeout,
            **self._kwargs,
        )



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Enumeration",
]


class Enumeration(BaseCRM):
    """"""

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get field descriptions for custom field type

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-fields/crm-userfield-enumeration-fields.html

        The method returns the field descriptions for a custom field of type 'enumeration' (list).

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)



import typing

from ..utils import enum as _enum
from ..utils import types as _types

__all__ = [
    "DEFAULT_INITIAL_RETRY_DELAY",
    "DEFAULT_MAX_RETRIES",
    "DEFAULT_RETRY_DELAY_INCREMENT",
    "DEFAULT_TIMEOUT",
    "B24AppStatus",
    "B24BoolLit",
]


DEFAULT_INITIAL_RETRY_DELAY: typing.Final[_types.Number] = 1
""""""

DEFAULT_MAX_RETRIES: typing.Final[int] = 3
""""""

DEFAULT_RETRY_DELAY_INCREMENT: typing.Final[_types.Number] = 0
""""""

DEFAULT_TIMEOUT: typing.Final[_types.DefaultTimeout] = 10
""""""


class B24AppStatus(_enum.StrEnum):
    """"""

    FREE = "F"
    DEMO = "D"
    TRIAL = "T"
    PAID = "P"
    LOCAL = "L"
    SUBSCRIPTION = "S"

    @property
    def is_free(self) -> bool:
        return self == self.FREE

    @property
    def is_demo(self) -> bool:
        return self == self.DEMO

    @property
    def is_trial(self) -> bool:
        return self == self.TRIAL

    @property
    def is_paid(self) -> bool:
        return self == self.PAID

    @property
    def is_local(self) -> bool:
        return self == self.LOCAL

    @property
    def is_subscribtion(self) -> bool:
        return self == self.SUBSCRIPTION


class B24BoolLit(_enum.StrEnum):
    """"""

    TRUE = "Y"
    FALSE = "N"
    DEFAULT = "D"

    def __bool__(self):
        return bool(_types.B24Bool(typing.cast(_types.B24BoolLiteral, self)))



from typing import Optional, Text

from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import type_checker
from ..utils.types import JSONDict, Timeout
from ._base_scope import BaseScope

__all__ = [
    "Placement",
]


class Placement(BaseScope):
    """"""

    @type_checker
    def bind(
            self,
            placement: Text,
            handler: Text,
            *,
            title: Optional[Text] = None,
            description: Optional[Text] = None,
            group_name: Optional[Text] = None,
            lang_all: Optional[JSONDict] = None,
            options: Optional[JSONDict] = None,
            user_id: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "PLACEMENT": placement,
            "HANDLER": handler,
        }

        if title is not None:
            params["TITLE"] = title

        if description is not None:
            params["DESCRIPTION"] = description

        if group_name is not None:
            params["GROUP_NAME"] = group_name

        if lang_all is not None:
            params["LANG_ALL"] = lang_all

        if options is not None:
            params["OPTIONS"] = options

        if user_id is not None:
            params["USER_ID"] = user_id

        return self._make_bitrix_api_request(
            api_wrapper=self.bind,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            scope: Optional[Text] = None,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = dict()

        if scope is not None:
            params["SCOPE"] = scope

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def unbind(
            self,
            placement: Text,
            handler: Optional[Text] = None,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = {
            "PLACEMENT": placement,
        }

        if handler is not None:
            params["HANDLER"] = handler

        return self._make_bitrix_api_request(
            api_wrapper=self.unbind,
            params=params,
            timeout=timeout,
        )



from typing import Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Trigger",
]


class Trigger(BaseCRM):
    """The class provide methods for working with application triggers.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/automation/triggers/index.html
    """

    @type_checker
    def __call__(
            self,
            *,
            target: Text,
            code: Optional[Text] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Activate trigger.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automation/crm-automation-trigger.html

        The method activates the webhook trigger configured in CRM automation.

        Args:
            target: Target object for automation, specified in the form of TYPENAME_ID;

            code: Unique symbolic code of the trigger configured in Automation for a specific statux/stage of the document;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "target": target,
        }

        if code is not None:
            params["code"] = code

        return self._make_bitrix_api_request(
            api_wrapper=self,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            *,
            code: Text,
            name: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add trigger.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automation/triggers/crm-automation-trigger-add.html

        The method adds a trigger.

        Args:
            code: Internal unique identifier of the trigger;

            name: Name of the trigger;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "CODE": code,
            "name": name,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def execute(
            self,
            *,
            code: Text,
            owner_type_id: int,
            owner_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Executes the trigger.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automation/triggers/crm-automation-trigger-execute.html

        The method triggers the execution of a trigger.

        Args:
            code: Internal unique identifier of the trigger;

            owner_type_id: Type of the CRM object;

            owner_id: Identifier of the entity;

            timeout: Timeout in seconds.

        Returns:
            Instance on BitrixAPIRequest
        """

        params = {
            "CODE": code,
            "OWNER_TYPE_ID": owner_type_id,
            "OWNER_ID": owner_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.execute,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of triggers.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automation/triggers/crm-automation-trigger-list.html

        The method retrieves a list of application triggers.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(timeout=timeout)

    @type_checker
    def delete(
            self,
            *,
            code: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete triggers.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/automation/triggers/crm-automation-trigger-delete.html

        The method deletes a trigger.

        Args:
            code: Internal unique identifier of the trigger;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "CODE": code,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Optional

from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.functional import type_checker
from .....utils.types import JSONDict
from ..._base_crm import BaseCRM
from .delivery import Delivery
from .product import Product

__all__ = [
    "Payment",
]


class Payment(BaseCRM):
    """The methods provide capabilities for managing payments.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/index.html
    """

    @cached_property
    def delivery(self) -> Delivery:
        """"""
        return Delivery(self)

    @cached_property
    def product(self) -> Product:
        """"""
        return Product(self)

    @type_checker
    def add(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Add payment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-add.html

        This method creates a payment for a CRM object.

        Args:
            entity_type_id: Identifier of the CRM object type;

            entity_id: Identifier of the CRM object;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Get payment by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-get.html

        This method retrieves brief information about the payment.

        Args:
            bitrix_id: Payment identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Get a list of payments.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-list.html

        Args:
            entity_type_id: Identifier of the CRM entity type;

            entity_id: Identifier of the CRM entity;

            filter: Object format:

                {
                    "field_1": "value_1",

                    "field_2": "value_2",

                    ...,

                    "field_n": "value_n",
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        if filter is not None:
            params["filter"] = filter

        if order is not None:
            params["order"] = order

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Update payment fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-update.html

        This method updates a limited set of payment fields.

        Args:
            bitrix_id: Payment identifier;

            fields: Object format:
                {
                    "paid": payment status, where "Y" for paid and "N" is not paid,

                    "paySystemId":  Payment system identifier
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(bitrix_id, fields, timeout=timeout)

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Delete payment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-delete.html

        This method deletes a payment.

        Args:
            bitrix_id: Payment identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)

    @type_checker
    def pay(
            self,
            bitrix_id: int,
            *,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Change payment status to "Paid".

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-pay.html

        This method changes the payment status to "Paid".

        Args:
            bitrix_id: Payment identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.pay,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def unpay(
            self,
            bitrix_id: int,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Change payment status to "Unpaid".

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/crm-item-payment-unpay.html

        This method changes the payment status to "Unpaid".

        Args:
            bitrix_id: Payment identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.unpay,
            params=params,
            timeout=timeout,
        )



from functools import cached_property

from .._base_crm import BaseCRM
from ._images import Icon, Logo
from .bindings import Bindings
from .comment import Comment
from .logmessage import Logmessage
from .note import Note

__all__ = [
    "Timeline",
]


class Timeline(BaseCRM):
    """"""

    @cached_property
    def bindings(self) -> Bindings:
        """"""
        return Bindings(self)

    @cached_property
    def comment(self) -> Comment:
        """"""
        return Comment(self)

    @cached_property
    def icon(self) -> Icon:
        """"""
        return Icon(self)

    @cached_property
    def logmessage(self) -> Logmessage:
        """"""
        return Logmessage(self)

    @cached_property
    def logo(self) -> Logo:
        """"""
        return Logo(self)

    @cached_property
    def note(self) -> Note:
        """"""
        return Note(self)



from typing import Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Note",
]


class Note(BaseCRM):
    """The methods provide capabilities for working with Notes in timeline.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/note/index.html
    """

    @type_checker
    def get(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            item_type: int,
            item_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get Information about note.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/note/crm-timeline-note-get.html

        This method returns information about a note related to a timeline record.

        Args:
            entity_type_id: Identifier of the entity type to which the record belongs;

            entity_id: Identifier of the entity to which the record belongs;

            item_type: Type of the record to which the note should be applied:

                - 1 — history record
                - 2 — deal;

            item_id: Identifier of the record to which the note should be applied;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "itemType": item_type,
            "itemId": item_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            item_type: int,
            item_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete note.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/note/crm-timeline-note-delete.html

        Args:
            entity_type_id: Identifier of the entity type to which the record belongs;

            entity_id: Identifier of the entity to which the record belongs;

            item_type: Type of the record to which the note should be applied:

                - 1 — history record
                - 2 — deal;

            item_id: Identifier of the record to which the note should be applied;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "itemType": item_type,
            "itemId": item_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def save(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            item_type: int,
            item_id: int,
            text: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Save note.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/note/crm-timeline-note-save.html

        Args:
            entity_type_id: Identifier of the entity type to which the record belongs;

            entity_id: Identifier of the entity to which the record belongs;

            item_type: Type of the record to which the note should be applied:

                - 1 — history record
                - 2 — deal;

            item_id: Identifier of the record to which the note should be applied;

            text: Note text;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "itemType": item_type,
            "itemId": item_id,
            "text": text,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.save,
            params=params,
            timeout=timeout,
        )



from typing import Any, Dict

from b24pysdk.bitrix_api.requests import BitrixAPIRequest


class DummyToken:
    def __init__(self, response: Dict[str, Any]):
        self._response = response
        self.calls = []

    def call_method(self, api_method, params=None, *, timeout=None, **kwargs):
        self.calls.append((api_method, params, timeout, kwargs))
        return self._response


def make_json_response(result: Any) -> Dict[str, Any]:
    return {
        "result": result,
        "time": {
            "start": 0.0,
            "finish": 0.1,
            "duration": 0.1,
            "processing": 0.05,
            "date_start": "2024-01-01T00:00:00+00:00",
            "date_finish": "2024-01-01T00:00:00+00:00",
        },
    }


def test_bitrix_api_request_deferred_call_and_result():
    token = DummyToken(make_json_response({"a": 1}))
    req = BitrixAPIRequest(bitrix_token=token, api_method="crm.deal.get", params={"bitrix_id": 1})

    # Deferred: not called until result/time accessed
    assert token.calls == []

    # Access result triggers call
    assert req.result == {"a": 1}
    assert len(token.calls) == 1

    # Accessing time uses cached responses, no extra call
    _ = req.time
    assert len(token.calls) == 1

    # __str__ contains method and params
    s = str(req)
    assert "crm.deal.get" in s
    assert "bitrix_id=1" in s



from functools import cached_property
from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM
from .badge import Badge
from .binding import Binding
from .communication import Communication
from .configurable import Configurable
from .layout import Layout
from .todo import Todo
from .type import Type

__all__ = [
    "Activity",
]


class Activity(BaseCRM):
    """Methods for working with system activities in the timeline.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/index.html
    """

    @cached_property
    def badge(self) -> Badge:
        """"""
        return Badge(self)

    @cached_property
    def binding(self) -> Binding:
        """"""
        return Binding(self)

    @cached_property
    def communication(self) -> Communication:
        """"""
        return Communication(self)

    @cached_property
    def configurable(self) -> Configurable:
        """"""
        return Configurable(self)

    @cached_property
    def layout(self) -> Layout:
        """"""
        return Layout(self)

    @cached_property
    def todo(self) -> Todo:
        """"""
        return Todo(self)

    @cached_property
    def type(self) -> Type:
        """"""
        return Type(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get activity fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-fields.html

        The method returns a description of the fields of the system activity.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new activity

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-add.html

        The method creates a new system activity.

        fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get activity by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-get.html

        The method returns information about the activity by its ID.

        Args:
            bitrix_id: Identifier of the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of activities.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-list.html

        The method returns a list of activities based on the filter, considering the permissions of the current user.

        Args:
            select: List of fields that should be populated in the selected elements;

            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-update.html

        The method updates an existing activity.

        Args:
            bitrix_id: Identifier of the activity to be changed;

            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-delete.html

        The method removes an activity of any type.

        Args:
            bitrix_id: Identifier of the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)



from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._base_crm import BaseCRM

__all__ = [
    "Vat",
]


class Vat(BaseCRM):
    """The methods provide capabilities for managing VAT rates in trade catalog.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/index.html
    """

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create VAT rate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/crm-vat-add.html

        The method creates a new VAT rate in CRM.

        Args:
            fields: Object format:
                {
                    field_1: value_1,
                    field_2: value_2,
                    ...,
                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete VAT rate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/crm-vat-delete.html

        The method removes a VAT rate by its identifier.

        Args:
            bitrix_id: Identifier of the VAT rate to be deleted;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get VAT rate by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/crm-vat-get.html

        The method returns the VAT rate parameters by ID.

        Args:
            bitrix_id: Identifier of the VAT rate to be deleted;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get VAT rate fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/crm-vat-fields.html

        The method returns the description of VAT rate fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            order: Optional[JSONDict] = None,
            filter: Optional[JSONDict] = None,
            select: Optional[Iterable[Text]] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of VAT rates by filter.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/crm-vat-list.html

        The method returns a list of VAT rates based on the filter.

        Args:
            order: Object format:
                {
                    field_1: value_1,
                    field_2: value_2,
                    ...,
                    field_n: value_n,
                },

                where

                    - field_n is the name of the field by which the selection will be sorted

                    - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            filter: Object format:
                {
                    field_1: value_1,
                    field_2: value_2,
                    ...,
                    field_n: value_n,
                };

            select: Array of returned fields;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update existing VAT rate.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/vat/crm-vat-update.html

        The method updates the parameters of an existing VAT rate.

        Args:
            bitrix_id: Identifier of the VAT rate to be updated;

            fields: Array of fields to update;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id=bitrix_id,
            fields=fields,
            timeout=timeout,
        )



from abc import ABC
from typing import TYPE_CHECKING

from ._base_context import BaseContext

if TYPE_CHECKING:
    from .. import Client


class BaseScope(BaseContext, ABC):
    """"""

    __slots__ = ("_client",)

    _client: "Client"

    def __init__(self, client: "Client"):
        self._client = client

    def __repr__(self):
        return f"scopes.{self.__class__.__name__}(client={self._client})"

    @property
    def _context(self) -> "Client":
        """"""
        return self._client
